% Script to test the JBD-NCG algorithm of the paper
% D. Nion, "A Tensor Framework for Non-Unitary Joint Block Diagonalization",
% IEEE Trans. on Signal Processing, Vol. 59, No.10, pp. 4585-4594, Oct. 2011.

clear all
close all
clc
%*******************************
% DIMENSIONS
%*******************************

%%------- Example 1: JBD under-determined case (A is fat), noise-free model, complex data, hermitian symetry 
% L_vec=[2 2 2 2];  % size of the blocks
% I=7;            % number of rows of the matrix A
% K=30;           % number of matrices to block-diagonalize
% SNR=inf;        % Signal To Noise Ratio (Set to inf to build an exact JBD problem)
% data_type='complex';       % 'complex or 'real'
% symf='herm';            % 'herm' for hermitian symetry or 'trans' for standard symetry

%%------- Example 2: JBD exactly-determined case (A is square), noisy data, real data
% L_vec=[3 3 3];  % size of the blocks
% I=9;            % number of rows of the matrix A
% K=30;           % number of matrices to block-diagonalize
% SNR=80;        % Signal To Noise Ratio (Set to inf to build an exact JBD problem)
% data_type='real';       % 'complex or 'real'
% symf='herm';            % 'herm' for hermitian symetry or 'trans' for standard symetry

% %%------- Example 3: JBD over-determined case (A is tall), noisy data, complex data, standard symetry
% L_vec=[3 3 3];  % size of the blocks
% I=11;            % number of rows of the matrix A
% K=30;           % number of matrices to block-diagonalize
% SNR=80;        % Signal To Noise Ratio (Set to inf to build an exact JBD problem)
% data_type='complex';       % 'complex or 'real'
% symf='trans';            % 'herm' for hermitian symetry or 'trans' for standard symetry

%%------- Example 4: JD (Joint-Diagonalization) over-determined case (A is tall), noisy data, real data
L_vec=[1 1 1 1 1];  % size of the blocks
I=7;                % number of rows of the matrix A
K=30;               % number of matrices to block-diagonalize
SNR=80;        % Signal To Noise Ratio (Set to inf to build an exact JBD problem)
data_type='real';       % 'complex or 'real'
symf='herm';            % 'herm' for hermitian symetry or 'trans' for standard symetry


R=length(L_vec);
N=sum(L_vec);
N2=sum(L_vec.^2);       
Nc=cumsum([0,L_vec]);
N2c=cumsum([0,L_vec.^2]);


%*******************
% ALGO PARAMETERS
%******************
Ninit=5;        % Number of initializations
Tol=1e-6;       % Tolerance
MaxIt=1000;     % Maximum number of iterations

%*********************************
% Generate Data
%*********************************
D=zeros(N,N,K);
if strcmp(data_type,'complex')==1
        A=randn(I,N)+j*randn(I,N);
        for r=1:R
           D(Nc(r)+1:Nc(r+1),Nc(r)+1:Nc(r+1),:)=randn(L_vec(r),L_vec(r),K)+j*randn(L_vec(r),L_vec(r),K); 
        end
else
        A=randn(I,N); 
        for r=1:R
           D(Nc(r)+1:Nc(r+1),Nc(r)+1:Nc(r+1),:)=randn(L_vec(r),L_vec(r),K); 
        end
end

% noise-free tensor (I by I by K) that collects the K matrices to block
% diagonalize across the 3rd dimension
if strcmp(symf,'trans')==1
    X=tmprod(tmprod(D,A,1),A,2);  
elseif strcmp(symf,'herm')==1
    X=tmprod(tmprod(D,A,1),conj(A),2); 
end
    
% add noise with symmetric slices
if strcmp(data_type,'complex')==1
    Noise_tens=randn(I,I,K)+j*randn(I,I,K);
else
    Noise_tens=randn(I,I,K);
end
if strcmp(symf,'herm')==1
    Noise_tens=Noise_tens+conj(permute(Noise_tens,[2 1 3]));
elseif strcmp(symf,'trans')==1
    Noise_tens=Noise_tens+permute(Noise_tens,[2 1 3]);
end
sigma=(10^(-SNR/20))*(norm(X(:),'fro')/norm(Noise_tens(:),'fro'));
X=X+sigma*Noise_tens;


%*******************************************************
% JBD algorithms
% compare:
% - JBD-SD (Steepest Descent), 
% - JBD-NCG with Polak-Ribire (PR) method
% - JBD-NCG with Hestenes-Stiefel (HS) method
% - JBD-NCG with Fletcher-Reeves (FR) method
% The first initialization is based on JBD closed form solution
% and the other initializations are random
%*************************************************************************
% Note 1: to use the algorithm with Ninit initializations generated inside
% the algorithm, use:
% [Ae,De,it,phi,phi_vec1]=JBD_NCG(X,L_vec,symf,'PR',Tol,MaxIt,Ninit);
% In the following, one wants to compare the different methods when they
% are all initialized by the same starting point, so instead we build a
% starting point Ainit outside the algorithm and we use
% [Ae,De,it,phi,phi_vec1]=JBD_NCG(X,L_vec,symf,'SD',Tol,MaxIt,[],Ainit);
%**************************************************************************
% Note 2:
% When we use
% [Ae,De,it,phi,phi_vec1]=JBD_NCG(X,L_vec,symf,'PR',Tol,MaxIt,Ninit);
% the Ninit initializations are tried (the first one is based on the closed form
% JBD solution and the others are random) until convergence is reached and
% the results obtained for the best initialization are given as outputs.
% However, it may not be needed to wait for convergence to be reached for
% each initialization. An alternative is to try Ninit initializations for a
% small number of iterations, and then to go on with the best
% initialization. In this case, the code below can be used. 
% First run, say 20, initializations with at most 10 iterations for each
% initialization
% MaxIt=10; 
% Ninit=20;
% [Ae,De,it1,phi,phi_vec1]=JBD_NCG(X,L_vec,symf,'PR',Tol,MaxIt,Ninit);
% Then go on with the best initialization only and with more iterations allowed:
% use the matrix Ae as input such that the algorithm restarts from this point
% MaxIt=1000; 
% [Ae,De,it2,phi,phi_vec2]=JBD_NCG(X,L_vec,symf,'PR',Tol,MaxIt,[],Ae);
% phi_vec=[phi_vec1,phi_vec2]
% it=it1+it2
%**************************************************************************

for ninit=1:Ninit
    % generate random initialization
    Ainit=randn(I,N)+j*randn(I,N);
    if strcmp(data_type,'real')==1
        Ainit=real(Ainit);
    end
    
disp('   ')
disp(['Press any key to run initialization ', num2str(ninit)])
disp('   ')
pause
    
disp('*************************************************************')
disp('ALGO 1 : Steepest descent (SD) algorithm')
disp('*************************************************************')
if ninit==1 % don't use Ainit, to use the exact JBD closed form solution to initialize
[Ae,De,it,phi,phi_vec1]=JBD_NCG(X,L_vec,symf,'SD',Tol,MaxIt,1);
else        % use Ainit
[Ae,De,it,phi,phi_vec1]=JBD_NCG(X,L_vec,symf,'SD',Tol,MaxIt,[],Ainit);
end
[MSE]=MSE_bdiag(Ae,A,L_vec,'subspace');
disp(['Results for initialization ',num2str(ninit),' :'])
disp(['Final value of the cost function:    ',num2str(phi)])
disp(['Iterations:                          ',num2str(it)])
disp(['MSE on A:                            ',num2str(MSE)])

disp('*************************************************************')
disp('ALGO 2 : JBD-NCG with Polak-Ribire (PR) method')
disp('*************************************************************')
if ninit==1 % don't use Ainit, to use the exact JBD closed form solution to initialize
[Ae,De,it,phi,phi_vec2]=JBD_NCG(X,L_vec,symf,'PR',Tol,MaxIt,1);
else        % use Ainit
[Ae,De,it,phi,phi_vec2]=JBD_NCG(X,L_vec,symf,'PR',Tol,MaxIt,[],Ainit);
end
[MSE]=MSE_bdiag(Ae,A,L_vec,'subspace');
disp(['Results for initialization ',num2str(ninit),' :'])
disp(['Final value of the cost function:    ',num2str(phi)])
disp(['Iterations:                          ',num2str(it)])
disp(['MSE on A:                            ',num2str(MSE)])

disp('*************************************************************')
disp('ALGO 3 : JBD-NCG with Hestenes-Stiefel (HS) method')
disp('*************************************************************')
if ninit==1 % don't use Ainit, to use the exact JBD closed form solution to initialize
[Ae,De,it,phi,phi_vec3]=JBD_NCG(X,L_vec,symf,'HS',Tol,MaxIt,1);
else        % use Ainit
[Ae,De,it,phi,phi_vec3]=JBD_NCG(X,L_vec,symf,'HS',Tol,MaxIt,[],Ainit);
end
[MSE]=MSE_bdiag(Ae,A,L_vec,'subspace');
disp(['Results for initialization ',num2str(ninit),' :'])
disp(['Final value of the cost function:    ',num2str(phi)])
disp(['Iterations:                          ',num2str(it)])
disp(['MSE on A:                            ',num2str(MSE)])

disp('*************************************************************')
disp('ALGO 4 : JBD-NCG with Fletcher-Reeves (FR) method')
disp('*************************************************************')
if ninit==1 % don't use Ainit, to use the exact JBD closed form solution to initialize
[Ae,De,it,phi,phi_vec4]=JBD_NCG(X,L_vec,symf,'FR',Tol,MaxIt,1);
else        % use Ainit
[Ae,De,it,phi,phi_vec4]=JBD_NCG(X,L_vec,symf,'FR',Tol,MaxIt,[],Ainit);
end
[MSE]=MSE_bdiag(Ae,A,L_vec,'subspace');
disp(['Results for initialization ',num2str(ninit),' :'])
disp(['Final value of the cost function:    ',num2str(phi)])
disp(['Iterations:                          ',num2str(it)])
disp(['MSE on A:                            ',num2str(MSE)])


%----------------------------------------------------------------
% PLOT CONVERGENCE CURVES for each initialization
%----------------------------------------------------------------
figure;
loglog(phi_vec1,'-b');hold on;
loglog(phi_vec2,'-r');hold on;      
loglog(phi_vec3,'-g');hold on;  
loglog(phi_vec4,'-magenta');hold on;  
legend('Steepest Descent','Polak-Ribire','Hestenes-Stiefel','Fletcher-Reeves')
title(['Evolution of the cost function for initialization ',num2str(ninit)])
xlabel('Number of iterations')
ylabel('Value of the cost function')
        
end