function [A,D,it,phi,phi_vec]=JBD_NCG(X,L_vec,symf,method,Tol,MaxIt,Ninit,A)
% Nonunitary Joint Block Diagonalization (JBD) by Non-linear Conjugate
% Gradient (NCG)
%--------------------------------------------------------------------------
% Author: Dimitri Nion (Feedback: dimitri.nion@gmail.com)
% @Copyright October 2011
% All rights reserved. This M-file and the code in it belongs to the holder 
% of the copyright. For non commercial use only.
%--------------------------------------------------------------------------
% If you make use of this code, please cite the following paper:
% D. Nion, "A Tensor Framework for Non-Unitary Joint Block Diagonalization",
% IEEE Trans. on Signal Processing, Vol. 59, No.10, pp. 4585-4594, Oct. 2011.
%--------------------------------------------------------------------------
%  JBD problem solved by this algorithm: 
%  Given a set of K matrices {X1,...,XK}, each of size IxI, 
%  estimate the non_unitary matrix A (I by N) and the N by N exactly 
%  block-diagonal matrices D1,...,DK such that:
% - if symf =='herm'  (hermitian symetry)
%           X1 = A * D1 * A'   
%			X2 = A * D2 * A'
%				     ...         
%			XK = A * DK * A'
% - if sym == 'trans' (standard symmetry)
%           X1 = A * D1 * A.'   
%			X2 = A * D2 * A.'
%				     ...          
%			XK = A * DK * A.'
%
%  A is not necessary tall: the under-determined case I<N is supported by
%  the algorithm.
%--------------------------------------------------------------------------
% The algorithm is based on the Nonlinear Conjugate Gradient  technique 
% with optimal step size applied on the cost function:
% - if symf =='herm'  (hermitian symetry)
%   phi = (1/2) norm(X - D x1 A x2 conj(A) , 'fro')^2
% - if sym == 'trans' (standard symmetry)
%   phi = (1/2) norm(X - D x1 A x2 A , 'fro')^2
% where xn is the mode-n tensor matrix product
%--------------------------------------------------------------------------
% INPUTS:  
% - X        :  IxIxK tensor, which holds the set of K matrices to block diagonalize
% - L_vec    :  =[L1, L2, ..., LR] size of the diagonal blocks (the way the partition is done)
% - symf     :  = 'trans' (if A.' in the second mode) 
%               = 'herm'  (if A' in the second mode)
%--------------------------------------------------------------------------
% OPTIONAL INPUTS:
% - method  (default 'PR')   : string to choose the method for updating the parameter beta of the conjugate direction
%                             'FR' for Fletcher-Reeves
%                             'PR' for Polak-Ribiere 
%                             'HS' for Hestenes-Stiefel
%                             'SD' for steepest descent (i.e., beta=0 such that the steepest descent direction is always enforced)
% - Tol   (default 1e-6)     : tolerance to stop the iterations
% - MaxIt (default 5000)     : maximum number of iterations
% - Ninit (default 5)        : number of initializations. The first one is built from the closed form solution to exact JBD problems, 
%                              and the Ninit-1 remainig initializations are random.
% - A                        : If provided, this matrix is used to initialize the algorithm. Then, Ninit is set to 1, i.e., the matrix 
%                              provided is the only initialization that is tried
%-------------------------------------------------------------------------------------------------------------------------------------------
% OUTPUTS: 
% - A       :  (IxN) estimate of the block-diagonalizer
% - D       :  (NxNxK) tensor which holds the exactly block-diagonal matrices, estimated in the LS sense from A
% - it      :  number of iterations
% - phi     :  final value of the cost function phi
% - phi_vec :  evolution of phi (vector of length it)
%-------------------------------------------------------------------------------------------------------------------------------------------
if nargin<3;error('At least 3 input arguments are required');end

%---------------------
% Size of the problem
%--------------------
I=size(X,1);
J=size(X,2);
K=size(X,3);
R=length(L_vec);
N=sum(L_vec);
N2=sum(L_vec.^2);       
Nc=cumsum([0,L_vec]);
N2c=cumsum([0,L_vec.^2]);

%-----------------
% Check inputs
%----------------
if I~=J;error('The first and second dimensions of X must be equal');end
if length(size(X))~=3;error('The input argument X has to be a third-order tensor');end
if exist('method')~=1 || isempty(method);method='PR';end
if exist('Tol')~=1 || isempty(Tol);Tol=1e-6;end
if exist('MaxIt')~=1 || isempty(MaxIt);MaxIt=5000;end
if exist('comp')~=1 || isempty(comp);comp='off';end
if exist('Ninit')~=1 || isempty(Ninit);Ninit=5;end

init_given=0;
if exist('A')==1 && isempty(A)~=1
    if (size(A,1)==size(X,1)) && (size(A,2)==N)    
        init_given=1; 
        A_init=A;  
        Ninit=1;       
    else
        error('The input matrix A provided to initialize the algorithm does not have the right dimensions')
    end
end

% matrix unfolding of X
X2=reshape(X,size(X,1)*size(X,2),size(X,3));  

%************************************************************************************************
%  Compute the decomposition for several starting points
%************************************************************************************************
phi_best=inf;              % Useful to select the best initialization

for ninit=1:Ninit   % Try Ninit different initializations
            
            %--------------------------------------------------------------------
            % BUILD INITIALIZATION
            %--------------------------------------------------------------------
           if init_given==0
                if ninit==1
                [A]= JBD_init(X,L_vec,'EVD',symf);
                else
                [A]=JBD_init(X,L_vec,'random',symf);  % force random init  
                end
                if strcmp(symf,'herm')==1
                        KAA=kat_rao_part(conj(A),A,L_vec,L_vec); 
                elseif strcmp(symf,'trans')==1
                        KAA=kat_rao_part(A,A,L_vec,L_vec); 
                end  
                 D_mat=KAA\X2;       % size N2 by K
                 % Build the corresponding tensor with block diagonal slices
                 D=zeros(N,N,K);
                 for r=1:R
                     D(Nc(r)+1:Nc(r+1),Nc(r)+1:Nc(r+1),:) = reshape(D_mat(N2c(r)+1:N2c(r+1),:),L_vec(r),L_vec(r),K);
                 end 
                  
                              
            elseif init_given==1 
                    A = A_init;
                    % Least squares update of D, given A
                    % D is an N2 by K matrix wheren L2 is varying more slowly than L1
                   if strcmp(symf,'herm')==1
                        KAA=kat_rao_part(conj(A),A,L_vec,L_vec); 
                    elseif strcmp(symf,'trans')==1
                        KAA=kat_rao_part(A,A,L_vec,L_vec); 
                    end     
                    D_mat=KAA\X2;       % size N2 by K   
                    % Build the corresponding tensor with block diagonal slices
                    D=zeros(N,N,K);
                    for r=1:R
                        D(Nc(r)+1:Nc(r+1),Nc(r)+1:Nc(r+1),:) = reshape(D_mat(N2c(r)+1:N2c(r+1),:),L_vec(r),L_vec(r),K);
                    end 

           end      

           
            %-------------------------------------------------------------------------
            % FIRST STEP in  STEEPEST DESCENT direction
            %-------------------------------------------------------------------------          
            % Initial value of phi
            Emat = X2 - KAA*D_mat;   % I^2 by K         
            phi=norm(Emat(:),'fro');        
            phi_vec=phi;
            it=1;
            stop=0;
            if MaxIt==0 || MaxIt<0
                stop=1;
            end

            %-----------------------------
            % Initial value of gradients
            %-----------------------------
            % Gradient wrt conj(D) 
            GD_mat = -0.5*KAA'*Emat;    % size N2xK
            % Initial value of search direction = steepest descent
            dD=-GD_mat; 
            % Gradient wrt conj(A)
            GA=zeros(I,N); 
            E=reshape(Emat,I,I,K);
               if strcmp(symf,'herm')==1
                  for k=1:K
                      GA = GA - E(:,:,k)'*A*D(:,:,k) - E(:,:,k)*A*D(:,:,k)';
                  end
                  GA=0.5*GA;
               elseif strcmp(symf,'trans')==1
                  for k=1:K
                      GA = GA - E(:,:,k).'*conj(A)*conj(D(:,:,k)) - E(:,:,k)*conj(A)*D(:,:,k)';
                  end
                  GA=0.5*GA;
               end
            % Initial value of search direction = steepest descent
            dA=-GA; 
            
            %----------------------------------------------------
            % Compute optimal steps in steepest descent direction
            %----------------------------------------------------
            [alphaA,alphaD]=Line_search_JBD(A,-GA,D_mat,-GD_mat,X2,L_vec,symf);

           %-----------------------------------------------
           % First update in the steepest descent direction
           %-----------------------------------------------
           A = A - alphaA*GA;
           D_mat = D_mat - alphaD*GD_mat;
                       
          
            %---- START MAIN LOOP --------------
                while stop==0

                        it=it+1;
                        phi_old=phi;
                              
                       %--------------------------------------------
                       % Compute Gradient wrt conj(D) and conj(A)
                       %--------------------------------------------
                       % Store old gradient values
                       GA_old=GA;
                       GD_mat_old=GD_mat;
                       
                       % Build the tensor D with block diagonal slices from D_mat
                        D=zeros(N,N,K);
                        for r=1:R
                            D(Nc(r)+1:Nc(r+1),Nc(r)+1:Nc(r+1),:) = reshape(D_mat(N2c(r)+1:N2c(r+1),:),L_vec(r),L_vec(r),K);
                        end
                        % compute kat_rao(A,A)
                        if strcmp(symf,'herm')==1
                          KAA = kat_rao_part(conj(A),A,L_vec,L_vec);  % I2I1xN2
                        elseif strcmp(symf,'trans')==1
                          KAA = kat_rao_part(A,A,L_vec,L_vec);
                        end
                        % Tensor of residual
                        Emat = X2 - KAA*D_mat;  
                        E=reshape(Emat,I,I,K);
                        
                       % Gradient wrt conj(D)
                       GD_mat = -0.5*KAA'*reshape(E,I*I,K);    % size N2xK
                      
                       % Gradient wrt conj(A)
                        GA=zeros(I,N); 
                        if strcmp(symf,'herm')==1
                            for k=1:K
                                GA = GA - E(:,:,k)'*A*D(:,:,k) - E(:,:,k)*A*D(:,:,k)';
                            end
                            GA=0.5*GA;
                        elseif strcmp(symf,'trans')==1
                            for k=1:K
                                GA = GA - E(:,:,k).'*conj(A)*conj(D(:,:,k)) - E(:,:,k)*conj(A)*D(:,:,k)';
                            end
                            GA=0.5*GA;
                        end
                        
                       %----------------------------------------------------------------
                       % Compute the restart parameter beta 
                       %------------------------------------------------------------------
                       NGA=norm(GA,'fro')^2;
                       NGD=norm(GD_mat,'fro')^2;
                       NGA_old=norm(GA_old,'fro')^2;
                       NGD_old=norm(GD_mat_old,'fro')^2;
                       diffGA = GA(:)-GA_old(:);
                       diffGD = GD_mat(:)-GD_mat_old(:);
                       if strcmp(method,'FR')==1         % Fletcher-Reeves 
                          betap = (NGA+NGD) / (NGA_old+NGD_old);
                       elseif strcmp(method,'PR')==1     % Polak-Ribiere
                          betap = (  real(GA(:)' * diffGA) + real(GD_mat(:)' * diffGD)  ) / (NGA_old+NGD_old);
                       elseif strcmp(method,'HS')==1     % Hestenes-Stiefel    
                          betap =  ( real(GA(:)' * diffGA) + real(GD_mat(:)' * diffGD)  ) / ( real(dA(:)' * diffGA) + real(dD(:)' * diffGD));
                       elseif strcmp(method,'SD')==1     % Steepest Descent
                          betap=0;
                       end
                       % restart strategy
                       betap=max([0,betap]);
                       if abs(real(GA(:)'*GA_old(:))+real(GD_mat(:)'*GD_mat_old(:)))/(NGA+NGD) > 0.1
                          betap=0;
                        end
                                                                       
                       %------------------------------------------
                       % Update conjugate direction
                       %------------------------------------------
                       dA = -GA     + betap * dA;
                       dD = -GD_mat + betap * dD;
                                               
                       %----------------------------------------
                       % Perform Line search
                       %----------------------------------------
                       [alphaA,alphaD]=Line_search_JBD(A,dA,D_mat,dD,X2,L_vec,symf);
                       
                       %----------------------------------------
                       % Update A and D
                       %----------------------------------------
                        A     =  A     + alphaA*dA;
                        D_mat =  D_mat + alphaD*dD;
                                                                      
                       %-----------------------------
                       % Stop criterion
                       %-----------------------------
                       phi=norm(E(:),'fro');
                       if (abs(phi-phi_old)/phi_old<Tol) || (it==MaxIt) || (phi<Tol)
                           stop=1;
                       end
                       if nargout>=5
                          phi_vec=[phi_vec, phi];
                       end
                                          
                                    
                end  % convergence reached for this initialization
                   
                    
                % Select this initialization if it is better than previous one
                     if (phi < phi_best)
                        phi_best=phi;
                        A_best=A;
                        D_best=D;
                        it_best=it;
                        phi_vec_best=phi_vec;
                        phi_vec=[];
                     end                  
                                   
end   % of all initializations
       
% Output arguments corresponding to the best initialization
A=A_best;
D=D_best;
it=it_best;
phi=phi_best;
phi_vec=phi_vec_best;

%*******************************************************************************************************
function Mat = kat_rao_part(B,C,partB,partC)
        % This function performs block wise kronecker product between the partitioned matrices B and C
        % Example: if B=[B1 B2 B3] and C=[C1 C2 C3], then M=[kron(B1,C1) kron(B2,C2) kron(B3,C3)]
        %---------------------------------------------------------------------------------------------
        % INPUTS:  
        % - B: of size JxM 
        % - C: of size KxN
        % - partB : vector that holds the number of columns of each block B1,B2,...
        % - partC : vector that holds the number of columns of each block C1,C2,...
        %   Example: partB=[L1 L2] means that M=L1+L2
        %-----------------------------------------------------------------------------------------------
        % OUTPUT
        % Mat of size JKxsum(partB.*partC) that contains the kronecker product of the blocks
        %------------------------------------------------------------------------------------------------
        
        [J M]=size(B);
        [K N]=size(C);
        if (sum(partB)~=M) 
            error(['Error: a matrix with ',int2str(M),' columns can not be partitioned in such a way'])
        end
        if (sum(partC)~=N) 
            error(['Error: a matrix with ',int2str(N),' columns can not be partitioned in such a way'])
        end
        if length(partB)~=length(partC)
             error('Error: the 2 input matrices do not have the same number of blocks')
        end
        
        indB=[0 cumsum(partB)];
        indC=[0 cumsum(partC)];
        indMat=[0 cumsum(partB.*partC)];

         Mat=zeros(J*K,sum(partB.*partC));
         for i=1:length(partC)  
             Mat(:,indMat(i)+1:indMat(i+1))=fast_kron( B(:,indB(i)+1:indB(i+1)) , C(:,indC(i)+1:indC(i+1)));
         end
         
%*********************************************************************************         
function C = fast_kron (A,B)

        %--------------------------------------------------------------------------
        % Compute the Kronecker Product of A and B, equivalent to kron.m function
        %--------------------------------------------------------------------------
        [I,L1]=size(A);
        [J,L2]=size(B);

        if (L1==1) && (L2==1)
             C=reshape(B*A.',I*J,1);
        elseif (L1==1) && (L2>1)
            Bt=B.'; 
            C=reshape(Bt(:)*A.',L2,I*J).';
        elseif (L2==1) && (L1>1)
             C=reshape(B*A(:).',I*J,L1);
        else
             C=reshape(permute(reshape(B(:)*A(:).',[J,L2,I,L1]),[1 3 2 4]),[I*J,L1*L2]);
        end
                                      
%************************************************************************************************
function X_out = tmprod(X,U,mode)
        %----------------------------------------------------------------------------------
        % This function computes the mode-n product between a 3rd order tensor and a matrix
        %----------------------------------------------------------------------------------
        % INPUTS: - tensor X of size (IxJxK)
        %         - Matrix U with I columns (if mode==1) or J columns (if mode==2) or K columns (if mode==3)
        %         - mode= 1, 2 or 3
        %----------------------------------------------------------------------------------
        % OUTPUTS: X_out results from mode-n (n=1,2 or 3) product between X and U
        %----------------------------------------------------------------------------------
        [I,J,K]=size(X);
        [M,N]=size(U);
        
        if (mode~=1) & (mode~=2) & (mode~=3)
            error('The input variable mode should be 1, 2 or 3')
        end
        
        if N~=size(X,mode) 
            error(['The number of columns of the input matrix should be equal to dimension ',int2str(mode),' of the input tensor'])
        end    

        if mode==1
            X_out = reshape(U*reshape(X,I,J*K) ,M,J,K);      
        elseif mode==2
            X_out = permute(reshape (U*reshape(permute(X,[2 1 3]),J,I*K), M,I,K),[2 1 3]);        
        elseif mode==3
            X_out = permute(reshape (U*reshape(permute(X,[3 1 2]),K,I*J), M,I,J),[2 3 1]);
        end
        
  %********************************************************************************************         
function [A,D]= JBD_init(X,L_vec,init_type,symf)     
        % INITIALIZATION FOR JOINT BLOCK DIAGONALIZATION ALGORITHMS
        %-------------------------------------------------------------------------------------------------------
        % Build an initialization based on the JBD closed form solution or a random initialization
        %--------------------------------------------------------------------------------------------------------
        % Author: Dimitri Nion, May 2010
        %-----------------------------------------------------------------------------------------------------------------------------------
        % INPUTS: 
        % - X (IxIxK)           :  Tensor which holds a set of K matrices of the form
        %                           X(:,:,k) = A * D(:,:,k) * A.' + Noise  if symf=='trans',
        %                           X(:,:,k) = A * D(:,:,k) * A'  + Noise  if symf=='herm',
        %                         where D(:,:,k) is a block-diagonal matrix.
        % - L_vec= [L1,...,LR]  :  holds the partition of the matrices Ar, r=1...R
        %                          i.e. Ar is of size IxLr and A=[A1,...,AR] of size IxN where N=sum(L_vec)
        % - symf                : either 'herm' or 'trans'
        % - init_type           : either 'random'  --> then A and D are built randomly 
        %                           or   'EVD'     --> then A and D are found from X by a closed form solution from EVD-ESPRIT like idea
        % - meth                : method used to compute the closed form solution
        %----------------------------------------------------------------------------------------------------------------------------------
        % OUTPUTS:
        % A: matrix of size IxN
        % D: tensor of size NxNxK with nearly block-diagonal slices
        %---------------------------------------------------------------------------------------------------------
        % Note:  The closed form solution (i.e. init_type='EVD') can only be computed if I>=N
        %        If I<N, then a random starting point is built
        %---------------------------------------------------------------------------------------------------------
     
        %---------------------------
        % Dimensions of the problem
        %---------------------------
        R=length(L_vec);
        N=sum(L_vec);
        I=size(X,1);
        K=size(X,3); 
        Nc=cumsum([0,L_vec]);
        N2c=cumsum([0,L_vec.^2]);
        if (I<N); init_type='random';end
        
        %-------------------------------
        % Build the starting point A
        %-------------------------------
        if strcmp(init_type,'EVD')==1   % closed form solution
            [A,D] = JBD_exact(X,L_vec,symf);
                 
        elseif  strcmp(init_type,'random')==1  % random initialization
            % generate random matrix A
            A=randn(I,N)+j*randn(I,N);
            if isreal(X);A=real(A);end
            % Orthogonalize A or its submatrices
            if I>=N;A=orth(A);
            else for r=1:R;A(:,Nc(r)+1:Nc(r+1))=orth(A(:,Nc(r)+1:Nc(r+1)));end;end          
              
            %------------------------------------------------------
            % Given A, build D in the least squares sense
            %------------------------------------------------------
            if (strcmp(symf,'trans')==1) || isreal(X)
               DD = kat_rao_part(A,A,L_vec,L_vec)\reshape(X,I*I,K);     % of size N2xK;
            elseif strcmp(symf,'herm')==1
               DD = kat_rao_part(conj(A),A,L_vec,L_vec)\reshape(X,I*I,K);     % of size N2xK;
            end
            % restack into a tensor with block diagonal slices
            D=zeros(N,N,K);
            for r=1:R
                D((Nc(r)+1):Nc(r+1),(Nc(r)+1):Nc(r+1),:)=reshape( DD(N2c(r)+1:N2c(r+1),:) , L_vec(r), L_vec(r), K);
            end

        end

%*****************************************************************************************************************************************
function [A,D,B] = JBD_exact(X,L_vec,symf)
        % CLOSED FORM SOLUTION FOR EXACT JOINT-BLOCK-DIAGONALIZATION (JBD) PROBLEMS
        %-----------------------------------------------------------------------------------------------------------------------------------
        % This function allows to solve exact JBD problems in a non-iterative way. In presence of noise, it can be used as a good 
        % initialization of optimization algorithms for the Approximate Joint Block Diagonalization problem.
        %-----------------------------------------------------------------------------------------------------------------------------------
        % JBD FORMULATION
        % Given a set of IxI matrices {Xk, k=1,...,K},
        % find the IxN full column rank matrix A and the set of NxN block diagonal matrices {Dk, k=1,...,K},
        % such that 
        % - if symf=='herm'   (hermitian symmetry)
        %         Xk, k=1,...,K can be approximated by 
        %                   Xk = A * Dk *A' 
        %   or equivalently 
        %                   Dk = B * Xk *B'
        %   where B=pinv(A) is NxI 
        % - if symf=='trans'  (symmetry)
        %         Xk, k=1,...,K can be approximated by 
        %                   Xk = A * Dk *A.' 
        %         or equivalently 
        %                   Dk = B * Xk *B.'
        %   where B=pinv(A) is NxI
        % In tensor format this can be written as follows:
        % X is the IxIxK tensor that holds the K matrices Xk and D is the NxNxK tensor that holds the K block-diagonal matrices Dk,
        % then we have
        %           X = D x1 A x2 conj(A) if symf='herm' or  X = D x1 A x2 A  if symf='trans'
        %           D = X x1 B x2 conj(B) if symf='herm' or  D = X x1 B x2 B  if symf='trans'
        % where x1 and x2 are the mode-1 and mode-2 tensor-matrix products, respectively.
        %-----------------------------------------------------------------------------------------------------------------------------------
        % INPUTS: 
        % - X (IxIxK)           :  Tensor which holds a set of K matrices of the form
        %                           X(:,:,k) = A * D(:,:,k) * A.' + Noise  if symf=='trans',
        %                           X(:,:,k) = A * D(:,:,k) * A'  + Noise  if symf=='herm',
        %                         where D(:,:,k) is a block-diagonal matrix.
        % - L_vec= [L1,...,LR]  :  holds the partition of the matrices Ar, r=1...R
        %                         i.e. Ar is of size IxLr and A=[A1,...,AR] of size IxN where N=sum(L_vec)
        % - symf                : either 'herm' or 'trans'
        %-----------------------------------------------------------------------------------------------------------------------------------
        % OUTPUTS:
        % - matrix A (IxN)
        % - tensor D (NxNxK), of which each of the K slices is approximately (not exactly) block diagonal
        % - matrix B=pinv(A) (NxI)
        %-----------------------------------------------------------------------------------------------------------------------------------
       
        % Dimensions of the problem
        N=sum(L_vec);
        I=size(X,1);
        K=size(X,3);
        N2=sum(L_vec.^2); 

        % SVD in third mode: rank-(I,I,Ns) approximation of X  ---> use svd
        [U,S,V]=svd(reshape(X,I^2,K),'econ');
        Ns=2;  % to select the Ns most significant slices
        X_core=reshape(U(:,1:min(N2,K))*S(1:min(N2,K),1:min(N2,K))*V(1:Ns,1:min(N2,K))',I,I,Ns);  
        % Generalized EVD of the first 2 slices of X_core
        X1=X_core(:,:,1);
        X2=X_core(:,:,2);
        [u1,s1,v1]=svd(X1,0);
        u1=u1(:,1:N);s1=s1(1:N,1:N);v1=v1(:,1:N);
        [u2,s2,v2]=svd(X2,0);
        u2=u2(:,1:N);s2=s2(1:N,1:N);v2=v2(:,1:N);
        X12=u1*s1*v1'*v2*inv(s2)*u2';
        [E12 D12]=eig(X12);  
        D12=diag(D12);
        % Keep only the N most significant eigenvalues (because X1*pinv(X2) is I by I  and rank N<I)    
        [dd, ind]=sort(abs(D12),'descend');
        D12=D12(ind);
        E12=E12(:,ind);  % I by I
        
        % If X is real but E12 is complex (because D12 has one or several pairs of conjugate eigenvalues)
        % then transform the columns of A with a similarity matrix to get a real-valued one
         if isreal(D12)~=1 && isreal(X)==1  
             r=1;
             T=[0.5 -i*0.5;0.5 i*0.5];  % 2 by 2 block: similarity matrix
             while r<=I
                  if isreal(D12(r))~=1
                      E12(:,r:r+1)= E12(:,r:r+1)*diag(exp(-i*sum(angle(E12(1,r:r+1)))/2))*T;   % Produces 2 real columns
                      r=r+2;
                  else
                      r=r+1;
                  end
             end 
             E12=real(E12);
         end
        E12=E12(:,1:N);
        B=pinv(E12);
        
        % find permutation with all slices 
        if strcmp(symf,'trans')==1
           T=tmprod(tmprod(X,B,1),B,2);
        elseif strcmp(symf,'herm')==1
           T=tmprod(tmprod(X,B,1),conj(B),2);
        end
        % each slice of T is a permuted block diagonal matrix, with the same permutation across all slices
        % The N*N matrix where each entry is the norm of the mode-3 vectors and find the permutation
        T_mat=(1/K)*sum(abs(T),3);
        [P12]=find_permblock(T_mat,L_vec);
        A=E12*P12;
        B=pinv(A);
        if strcmp(symf,'trans')==1
           D=tmprod(tmprod(X,B,1),B,2);
        elseif strcmp(symf,'herm')==1
           D=tmprod(tmprod(X,B,1),conj(B),2);
        end

%**********************************************************************************************************************************
function [P,D]=find_permblock(Q,L_vec)
% Estimation of permutation matrix that allows to recover a block-diagonal structure
%------------------------------------------------------------------------------------------------------------------
% INPUTS: 
% - Q (NxN)  :  block diagonal square matrix of which rows and columns have been permuted
% - L_vec    : =[L1,L2,...,LR], N=sum(L_vec), size of each of the R square blocks
%------------------------------------------------------------------------------------------------------------------
% OUTPUTS: 
% - P (NxN)      : permutation matrix such that Q=P*D*P.'
% - D=P.'*Q*P    : is as most block diagonal as possible, with bloc(1,1) of size L1xL1, bloc(2,2) of size L2xL2 etc 
%------------------------------------------------------------------------------------------------------------------
% Example:
% L_vec=[3 2 1 4 2];
% N=sum(L_vec);
% R=length(L_vec);
% Nc=[0,cumsum(L_vec)];
% D=zeros(N,N);
% for r=1:R; 
% D(Nc(r)+1:Nc(r+1),Nc(r)+1:Nc(r+1))=randn(L_vec(r),L_vec(r))+j*randn(L_vec(r),L_vec(r));
% end
% P=eye(N);
% P=P(:,randperm(N));
% Q=P*D*P.';  % arbitrarily permute columns
% [P,D]=find_permblock(Q,L_vec);
%---------------------------------------------------------------------------------------------------------------------------


R=length(L_vec);
N=sum(L_vec);
V=diag(1./sqrt(diag(Q*Q')))*abs(Q);  % normalize each row of V1
Ind1=1:N;
indP1=[];

% if L_vec=[4 2 3 2 2 3 4 ],
% we first order L_vec=[2 2 2 3 3 4 4]
% then we first select the 2+2+2 positions where the sum of 2 elements is maximum
% and in the given set, we split into 3 groups of 2
% we delete the columns and rows and do the same for 3+3, and  4+4
[L_vec,pL]=sort(L_vec,'ascend');  
L_vec_asc=L_vec;

while length(L_vec)~=0
    Lr=L_vec(1);
    nb_Lr=length(find(L_vec==Lr));
    L_vec(1:nb_Lr)=[];
    % for each row of V, we compute the norm from the Lr greatest entries 
    N1=zeros(size(V,1),1);
    for i=1:size(V,1)
        vi=sort(V(i,:),'descend');
        N1(i,1)=norm(vi(1:Lr),'fro')^2;
    end
    % select the nb_Lr*Lr highest values
    [p1,indp1]=sort(N1,'descend');
    indp1=indp1(1:nb_Lr*Lr);  % indp1 holds the row indexes of V for the values Lr
    W=V(indp1,indp1);
    V(:,indp1)=[];
    V(indp1,:)=[];
    
    % we have selected the nb_Lr blocks of size Lr by Lr, now we have to reorder them
    IndL1=1:nb_Lr*Lr;
    indPL1=[];
    for i=1:nb_Lr
        x1=W(1,:);
        [x1,indx1]=sort(x1,'descend');
        indPL1=[indPL1  IndL1(indx1(1:Lr))];
        IndL1(indx1(1:Lr))=[];
        W(:,indx1(1:Lr))=[];
        W(indx1(1:Lr),:)=[];
    end

    indP1=[indP1  Ind1(indp1(indPL1.').')];
    Ind1(indp1(indPL1.').')=[];
        
end
   
% Since we had reordered L_vec in ascending order, reorder as in initial L_vec vector
Nc=[0,cumsum(L_vec_asc)];
L_cell=cell(1,R);
for r=1:R
    L_cell{r}=indP1(Nc(r)+1:Nc(r+1));
end
L_cell(:,pL)=L_cell;
indP1=cell2mat(L_cell);
P=eye(N,N);
P=P(:,indP1);
D=P.'*Q*P;

%**********************************************************************************************************************************
function [STEPA,STEPD,ZA,ZD]=Line_search_JBD(A,dA,D,dD,X,L_vec,symf)
% Computation of the optimal step (e.g. for gradient descent) for the computation of Joint Block Diagonalization problem
% - Given that A (IxN) is updated by   A = A + STEPA dA, where dA(IxN) is for instance equal to -gradient(A) 
% - Given that D (N2xK) is updated by  D = D + STEPD dD, where dD(N2xK) is for instance equal to -gradient(D)
% Find the optimal real valued steps STEPA and STEPD that minimize
%  phi = norm( (conj(A) + conj(STEPA)*conj(dA))(kr)(A + STEPA dA) * (D + STEPD dD)  - X  ,'fro' )^2 ,
% where X is the I^2 by K matrix representation of the observed tensor and (kr) is the block-wise Kronecker product
%-----------------------------------------------------------------------------------------------------------------------------
if strcmp(symf,'herm')==1
    KdAdA = kat_rao_part(conj(dA),dA,L_vec,L_vec);
    KAdA  = kat_rao_part(conj(A),dA,L_vec,L_vec);   
    KdAA  = kat_rao_part(conj(dA),A,L_vec,L_vec);
    KAA   = kat_rao_part(conj(A),A,L_vec,L_vec);
elseif strcmp(symf,'trans')==1
    KdAdA = kat_rao_part(dA,dA,L_vec,L_vec);
    KAdA  = kat_rao_part(A,dA,L_vec,L_vec);
    KdAA  = kat_rao_part(dA,A,L_vec,L_vec);
    KAA   = kat_rao_part(A,A,L_vec,L_vec);  
end
    

    M1=KdAdA*dD;
    M1=M1(:);
    M2=KdAdA*D;
    M2=M2(:);
    MM=KAdA+KdAA;
    M3=MM*dD;
    M3=M3(:);
    M4=MM*D;
    M4=M4(:);
    M5=KAA*dD;
    M5=M5(:);
    M6=KAA*D-X;
    M6=M6(:);

    % Matrices useful for updating STEPD
    Z1=[M2,M4,M6];
    Z2=[M1,M3,M5];
    Z12=Z2'*Z1;   % 3 by 3
    Z22=Z2'*Z2;   % 3 by  3
    Z11=Z1'*Z1;   % 3 by  3
    
    % Polynomials
    Q2=zeros(1,5);            % degree 4 polynomial in alphaA, coeffs of STEPD^2
    Q2(1)=real(Z22(1,1));
    Q2(2)=real(2*Z22(1,2));
    Q2(3)=real(2*Z22(1,3)+Z22(2,2));
    Q2(4)=real(2*Z22(3,2));
    Q2(5)=real(Z22(3,3));
    
    Q1=zeros(1,5);            % degree 4 polynomial in alphaA, coeffs of STEPD^1
    Q1(1)=real(Z12(1,1));
    Q1(2)=real(Z12(1,2)+Z12(2,1));
    Q1(3)=real(Z12(1,3)+Z12(3,1)+Z12(2,2));
    Q1(4)=real(Z12(3,2)+Z12(2,3));
    Q1(5)=real(Z12(3,3));
    
    
    Q0=zeros(1,5);            % degree 4 polynomial in alphaA, coeffs of STEPD^0
    Q0(1)=real(Z11(1,1));
    Q0(2)=real(2*Z11(1,2));
    Q0(3)=real(2*Z11(1,3)+Z11(2,2));
    Q0(4)=real(2*Z11(3,2));
    Q0(5)=real(Z11(3,3));
    
    Q0Q2=conv(Q0,Q2);
    Q1Q1=conv(real(Q1),real(Q1));
    polDnum=-Q1Q1+Q0Q2;   % degree 8
    polDnum_der=[8:-1:1].*polDnum(1:8);
    Q2_der=[4:-1:1].*Q2(1:4);
    sqrts=roots(conv(polDnum_der,Q2)-conv(polDnum,Q2_der));
    sqrts=sqrts(find(imag(sqrts)==0));
    % Find optimal STEP
    extremum=polyval(polDnum,sqrts)./polyval(Q2,sqrts);
    STEPA=sqrts(find(extremum==min(extremum),1));
    STEPA=STEPA(1);
       
   u=[STEPA^2 STEPA 1];
   STEPD=-real((u*Z12*u')/(u*Z22*u')); 
       
   
ZD=D+STEPD*dD;
ZA=A+STEPA*dA;
